{% extends 'generic/object_edit.html' %}
{% load form_helpers %}
{% load i18n %}

{% block title %}{% trans "Add Client Credential" %}{% endblock %}

{% block form %}
    {% render_form form %}
{% endblock %} 

{% block buttons %}
<a href="{{ return_url }}" class="btn btn-outline-secondary btn-float">{% trans "Cancel" %}</a>
  <div class="btn-group" role="group" aria-label="{% trans "Actions" %}">
    <button type="submit" name="_create" class="btn btn-primary">
      {% trans "Create" %}
    </button>
  </div>
{% endblock buttons %}
